//======================================================================
// 
// KeyTrap.c
//
// Written by Jon Daley for Carnegie-Mellon University, 05/27/1998
// Modified from Ctrl2Cap NTDDK sample
// Built and tested on WinNT 4.00.1381
//
// See comments in readme.txt for modifying code
//======================================================================

#include "ntddk.h"
#include "stdio.h"
#include "KeyTrap.h"
#include "ntddkbd.h"
 
// print macro that only turns on when checked builds are on
#if DBG
	#define DbgPrint(arg) DbgPrint arg
#else
	#define DbgPrint(arg)
#endif

#define NT_DEVICE_NAME L"\\Device\\KeyTrap"
#define DOS_DEVICE_NAME L"\\DosDevices\\KeyTrap"

ULONG on;  			// Trapping, on or off?
PDEVICE_OBJECT       kbdDevice, HookDeviceObject;
int F12counter; 	// Number of times F12 has been pressed in a row
int F12pos;			// Last Position of F12 key

/*
  //----------------------------------------------------------------------
  // 
  // KeyTrapUnload
  //
  // Unload keytrap driver
  // !!!!!*****  Problem *****!!!!!
  // Need two-step process to unload, this way will crash driver on next
  //    keystroke
  //
  // To "Unload" driver (disable keytrapping):
  //   close handle from win32 program...
  //
  //----------------------------------------------------------------------
VOID KeyTrapUnload(IN PDRIVER_OBJECT DriverObject){
UNICODE_STRING uniWin32NameString;
    DbgPrint(("Goodbye! -- Time to die..."));
    RtlInitUnicodeString(&uniWin32NameString, DOS_DEVICE_NAME);
    IoDeleteSymbolicLink(&uniWin32NameString);
    IoDeleteDevice(DriverObject->DeviceObject);
}
*/

  //----------------------------------------------------------------------
  // 
  // KeyTrapInit
  //
  // Initialize keytrap driver, print out debugging code
  //
  //----------------------------------------------------------------------
NTSTATUS KeyTrapInit( IN PDRIVER_OBJECT DriverObject ){
CCHAR		 	ntNameBuffer[64];
STRING		 	ntNameString;
UNICODE_STRING	ntUnicodeString;
NTSTATUS		status;

	sprintf(ntNameBuffer, "\\Device\\KeyboardClass0");
	RtlInitAnsiString(&ntNameString, ntNameBuffer);
	RtlAnsiStringToUnicodeString(&ntUnicodeString, &ntNameString, TRUE);
	status = IoCreateDevice(DriverObject,0,NULL,
				FILE_DEVICE_KEYBOARD,0,FALSE,&HookDeviceObject);
	if(!NT_SUCCESS(status)){
		DbgPrint(("**Error** Unable to create keyboard driver...\n"));
		RtlFreeUnicodeString(&ntUnicodeString);
		return STATUS_SUCCESS;
	}
	HookDeviceObject->Flags |= DO_BUFFERED_IO;
	status = IoAttachDevice(HookDeviceObject, &ntUnicodeString, &kbdDevice);
	if(!NT_SUCCESS(status)){
		DbgPrint(("**Error** Unable to attach to keyboard driver...\n"));
		IoDeleteDevice(HookDeviceObject);
		RtlFreeUnicodeString(&ntUnicodeString);
		return STATUS_SUCCESS;
	}
	RtlFreeUnicodeString(&ntUnicodeString);

	DbgPrint(("If you have any problems with this, delete")); 
	DbgPrint(("c:\\winnt\\system32\\drivers\\KeyTrap.sys"));
	DbgPrint(("-Jon Daley"));
	DbgPrint((""));
	DbgPrint(("Keyboard Mapper Driver Complete"));
	DbgPrint(("*****************************************"));

	return STATUS_SUCCESS;
}
  
  //----------------------------------------------------------------------
  // 
  // KeyTrapReadComplete
  //
  // Gets control after a read operation has completed.
  //
  //----------------------------------------------------------------------
NTSTATUS KeyTrapReadComplete( IN PDEVICE_OBJECT DeviceObject, IN PIRP Irp,
				 IN PVOID Context ){
PIO_STACK_LOCATION        IrpSp;
PKEYBOARD_INPUT_DATA      KeyData;
int                       numKeys, i;

	IrpSp = IoGetCurrentIrpStackLocation( Irp );
	if(NT_SUCCESS(Irp->IoStatus.Status)){
		KeyData = Irp->AssociatedIrp.SystemBuffer;
		numKeys = Irp->IoStatus.Information / sizeof(KEYBOARD_INPUT_DATA);
		for( i = 0; i < numKeys; i++ ) { 		// Step through all keys
			if(KeyData[i].MakeCode == F12 && 	// If F12 is active
			   KeyData[i].Flags == KEY_DOWN && 	// pushed down (not up)
			   F12pos == KEY_UP){ 	// If it was up previously 
			   						//   (not being held)
				 F12counter++;		// increment counter
				 F12pos = KEY_DOWN; // Record that key was previously down
			}
			else if(KeyData[i].MakeCode == F12 && 	// If F12 is active
					KeyData[i].Flags == KEY_UP)		// being released
				F12pos = KEY_UP;					// Record state of key
			else
				F12counter = 0;		// Not in a row, reset counter
			if(F12counter == 3){	// Reached F12, 3 times in a row
				on = !on;			// Switch Trapping state
				F12counter = 0;		// Reset counter for next time
				DbgPrint(("***** Trapping %s *****",
				                         on ? "on..." : "off..."));
			}
			if(on){ // Disable following keys
				if(KeyData[i].MakeCode == RIGHT_ALT ||
			 		KeyData[i].MakeCode == LEFT_ALT ||
//					KeyData[i].MakeCode == RIGHT_CTL ||
//					KeyData[i].MakeCode == LEFT_CTL ||
					KeyData[i].MakeCode == LEFT_WINDOWS ||
					KeyData[i].MakeCode == RIGHT_WINDOWS ||
					KeyData[i].MakeCode == APPLICATION ||
					KeyData[i].MakeCode == TAB ||
					KeyData[i].MakeCode == ESC ||
					KeyData[i].MakeCode == DEL){
						DbgPrint(("Disabled: %x - %s",
							KeyData[i].MakeCode, 
							KeyData[i].Flags ? "Up" : "Down"));
						KeyData[i].MakeCode = 0;
				}
				else // A non-disabled key
					DbgPrint(("Trapped: %x - %s",
						KeyData[i].MakeCode, 
						KeyData[i].Flags ? "Up" : "Down"));
			}   
			else // Monitoring keystrokes
				DbgPrint(("Scan: %x - %s\n",
				  KeyData[i].MakeCode, KeyData[i].Flags ? "Up" : "Down"));
		}	
	}
	if(Irp->PendingReturned)
		IoMarkIrpPending(Irp);
	return Irp->IoStatus.Status;
}

  //----------------------------------------------------------------------
  // 
  // KeyTrapOn
  //
  // Received CreateFile from win32 program, turn on trapping
  //
  //----------------------------------------------------------------------
NTSTATUS KeyTrapOn(IN PDEVICE_OBJECT DeviceObject,IN PIRP Irp){
	DbgPrint(("***** Trapping on *****"));
	on = 1;
    Irp->IoStatus.Status = STATUS_SUCCESS;
    Irp->IoStatus.Information = 0;
    IoCompleteRequest(Irp, IO_NO_INCREMENT);
    return STATUS_SUCCESS;
}

  //----------------------------------------------------------------------
  // 
  // KeyTrapOff
  //
  // Received CloseHandle from win32 program, turn off trapping
  //
  //----------------------------------------------------------------------
NTSTATUS KeyTrapOff(IN PDEVICE_OBJECT DeviceObject,IN PIRP Irp){
	DbgPrint(("***** Trapping off *****"));
	on = 0;
    Irp->IoStatus.Status = STATUS_SUCCESS;
    Irp->IoStatus.Information = 0;
    IoCompleteRequest(Irp, IO_NO_INCREMENT);
    return STATUS_SUCCESS;
}


//----------------------------------------------------------------------
//
// KeyTrapDispatchRead
//
// Sets up to look at the read request completion so that we can
// massage the input queue on IO completion.
//
//----------------------------------------------------------------------
NTSTATUS KeyTrapDispatchRead(IN PDEVICE_OBJECT DeviceObject, IN PIRP Irp){
PIO_STACK_LOCATION currentIrpStack = IoGetCurrentIrpStackLocation(Irp);
PIO_STACK_LOCATION nextIrpStack    = IoGetNextIrpStackLocation(Irp);
    // Push params down for keyboard class driver.
	*nextIrpStack = *currentIrpStack;
    // Set the completion callback, so we can "frob" the keyboard data.
	IoSetCompletionRoutine( Irp, KeyTrapReadComplete, 
							DeviceObject, TRUE, TRUE, TRUE );
    // Return the results of the call to the keyboard class driver.
	return IoCallDriver( kbdDevice, Irp );
}

//----------------------------------------------------------------------
//
// KeyTrapDispatchGeneral
//
// This handles several functions we are not interested in 
// along to the keyboard class driver. 
//
//----------------------------------------------------------------------
NTSTATUS KeyTrapDispatchGeneral(IN PDEVICE_OBJECT DeviceObject, IN PIRP Irp)
{
PIO_STACK_LOCATION currentIrpStack = IoGetCurrentIrpStackLocation(Irp);
PIO_STACK_LOCATION nextIrpStack    = IoGetNextIrpStackLocation(Irp);
	Irp->IoStatus.Status      = STATUS_SUCCESS;
	Irp->IoStatus.Information = 0;
// If this call was directed at the hook device, pass it to 
// the keyboard class device, else handle it ourselves.
	if(DeviceObject == HookDeviceObject){
		*nextIrpStack = *currentIrpStack;
		return IoCallDriver(kbdDevice, Irp);
	}
	return STATUS_SUCCESS;
}

NTSTATUS DriverEntry(IN PDRIVER_OBJECT  DriverObject, 
					IN PUNICODE_STRING RegistryPath){
PDEVICE_OBJECT	deviceObject = NULL;
NTSTATUS		ntStatus;
UNICODE_STRING 	uniNtNameString;
UNICODE_STRING 	uniWin32NameString;
UNICODE_STRING	deviceNameUnicodeString;

	DbgPrint(("*****************************************"));
	DbgPrint(("KeyTrap Driver Loading..."));

	RtlInitUnicodeString(&uniNtNameString, NT_DEVICE_NAME);
	ntStatus = IoCreateDevice(
		DriverObject,
		0,						// We don't use a device extension
		&uniNtNameString,		// Win32 Name
		FILE_DEVICE_KEYTRAP,	// Number I made up for this device
		0,						// No standard device characteristics
		FALSE,   				// This isn't an exclusive device--?
		&deviceObject);

	if(NT_SUCCESS(ntStatus)){
		DriverObject->MajorFunction[IRP_MJ_READ] = KeyTrapDispatchRead;
		DriverObject->MajorFunction[IRP_MJ_CREATE] = KeyTrapOn;
		DriverObject->MajorFunction[IRP_MJ_CLOSE] = KeyTrapOff;
//  Uncomment the following line if you have implemented the two-step
//  unloading process required for keyboard drivers
//		DriverObject->DriverUnload = KeyTrapUnload;
		DriverObject->MajorFunction[IRP_MJ_FLUSH_BUFFERS] = 
		DriverObject->MajorFunction[IRP_MJ_CLEANUP] = 
		DriverObject->MajorFunction[IRP_MJ_DEVICE_CONTROL] =
								KeyTrapDispatchGeneral;

		RtlInitUnicodeString(&uniWin32NameString, DOS_DEVICE_NAME);
		ntStatus = IoCreateSymbolicLink(&uniWin32NameString, 
										&uniNtNameString);
		if(!NT_SUCCESS(ntStatus)){
			DbgPrint(("**Error** Create Symbolic Link failed...\n"));
			IoDeleteDevice( DriverObject->DeviceObject );
		}
		else
			DbgPrint(("Initialized..."));
	}
	else
		DbgPrint(("Couldn't create the device..."));
  on = 0;		// Initialize counters etc. to 0...
  F12counter = 0;
  F12pos = KEY_UP;

  return KeyTrapInit(DriverObject);
}
